/**
 * Copyright (C) 2012 Yvon TANGUY
 * 
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.download;

import java.io.EOFException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.UnknownHostException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.zip.GZIPInputStream;

import org.vono.narau.Common;
import org.vono.narau.R;

import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.text.Html;
import android.util.Log;
import android.widget.TextView;

/**
 * Manage the databases downloads.
 */
public class DownloaderGzipDB {
	private static final String TAG = DownloaderGzipDB.class.getSimpleName();

	private static final String KEY_MSG_LOG_INFO = "linfo";
	private static final String KEY_MSG_LOG_WARN = "lwarn";
	private static final String KEY_MSG_LOG_ERROR = "lerror";

	/**
	 * Current download id (DBInfos)
	 */
	private static final String KEY_MSG_UPD_TOTAL = "pb_upd_total";
	/**
	 * Progress bar update.
	 */
	private static final String KEY_MSG_PB_UPD_VAL = "pb_upd_val";
	/**
	 * Download speed value.
	 */
	private static final String KEY_MSG_DNWL_SPEED = "speed";
	private static final String KEY_MSG_PB_ERROR = "pb_error";
	private static final String KEY_MSG_MENU = "show_menu";
	/**
	 * Message start download.
	 */
	private static final String KEY_MSG_START_DWNL = "start_dwnl";

	/**
	 * The update time in milliseconds of the downloads progress bar and the
	 * download speed text.
	 */
	private static final long UPDATE_TIME = 500;

	/**
	 * It is faster to do a multiplication than a division. So for an update
	 * every 0.5s, we need to multiply by 2 the receive bytes to get the value
	 * in bytes/s.
	 */
	private static final long SPEED_FACTOR = 2L;

	public static final int DWNL_INFO_SIZE = 0;
	public static final int DWNL_INFO_DWNL_SIZE = 1;
	public static final int DWNL_INFO_FILENAME = 2;

	private static final int BUFFER_SIZE = 1024 * 1024;

	private static Controller controller = null;
	private static UIHandler handler = null;
	private static boolean prematureStop;

	/**
	 * An input stream to be able to know how many bytes are reads from the
	 * inner input stream. Used to be able to update the progress bar.
	 * 
	 */
	private static class CounterInputStream extends InputStream {
		private InputStream in;

		private long readCount;
		private MessageDigest messageDigest;

		private CounterInputStream(InputStream in) {
			this.in = in;
			this.readCount = 0;

			try {
				this.messageDigest = MessageDigest.getInstance("MD5");
			} catch (NoSuchAlgorithmException nsae) {
				Log.e(TAG, "MD5 checksum: " + nsae);
			}
		}

		@Override
		public int read() throws IOException {
			int readByte = in.read();
			if (readByte != -1) {
				this.readCount++;
				if (null != this.messageDigest) {
					this.messageDigest.update((byte) readByte);
				}
			}

			return readByte;
		}

		@Override
		public int read(byte[] buffer, int offset, int length) throws IOException {

			int ret = this.in.read(buffer, offset, length);
			if (ret != -1) {
				this.readCount += ret;
				if (null != this.messageDigest) {
					this.messageDigest.update(buffer, offset, ret);
				}
			}
			return ret;
		}

		@Override
		public void close() throws IOException {
			if (this.in != null) {
				try {
					this.in.close();
				} finally {
					this.in = null;
				}
			}
		}

		private byte[] getMD5() {
			if (null != this.messageDigest) {
				return this.messageDigest.digest();
			}
			return null;
		}

	}

	/**
	 * Do the download job
	 * 
	 */
	private static class Downloader extends Thread {

		/**
		 * Used to build output as Hex
		 */
		private static final char[] DIGITS_LOWER = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c',
				'd', 'e', 'f' };

		private DBInfos dbInfos;
		private long downloadSize;
		private CounterInputStream cIS;
		private boolean success;

		private Downloader(DBInfos dbInfo) {
			this.cIS = null;
			this.success = false;
			this.dbInfos = dbInfo;
			this.resetDownloadSize();
			this.setName("Downloader " + dbInfo.name);
		}

		@Override
		public void run() {
			final DBInfos dbInfo = this.dbInfos;
			final File outFile = new File(Common.dataPath, dbInfo.dbName);
			final String downloadingMsg = Common.getString(R.string.downloadURL);
			final int numUrl = dbInfo.url.length;
			HttpURLConnection connection = null;

			final byte[] buffer = new byte[DownloaderGzipDB.BUFFER_SIZE];

			for (int i = 0; i < numUrl; i++) {
				// loop on the available urls

				GZIPInputStream gzIs = null;
				OutputStream fileOs = null;

				String message = String.format(downloadingMsg, dbInfo.url[i]);
				UIHandler.showMessage(KEY_MSG_LOG_INFO, message);

				try {
					URL url = dbInfo.url[i];

					connection = (HttpURLConnection) url.openConnection();
					connection.connect();

					// check the connect response code, should be in the 200
					int responseCode = connection.getResponseCode();
					if (responseCode / 100 != 2) {
						message = "" + responseCode + ": " + connection.getResponseMessage();
						UIHandler.showMessage(KEY_MSG_LOG_ERROR, message);

						connection.disconnect();
						connection = null;
						continue;
					}

					long dwnlSize = (long) connection.getContentLength();

					if (dwnlSize < 1) {
						UIHandler.showMessage(KEY_MSG_LOG_WARN, Common.getString(R.string.downloadWarnNoSize));
						dwnlSize = dbInfo.gzSize;
					}
					synchronized (TAG) {
						this.downloadSize = dwnlSize;
					}

					this.cIS = new CounterInputStream(connection.getInputStream());
					gzIs = new GZIPInputStream(this.cIS, DownloaderGzipDB.BUFFER_SIZE);

					fileOs = new FileOutputStream(outFile);

					int read = 0;
					while (read != -1) {
						read = gzIs.read(buffer, 0, DownloaderGzipDB.BUFFER_SIZE);

						if (read != -1) {
							fileOs.write(buffer, 0, read);
						}

						if (DownloaderGzipDB.isPrematureThreadStop()) {
							return;
						}
					}
					final byte[] md5Sum = this.cIS.getMD5();
					if (null != md5Sum) {

						final String md5download = new String(encodeHex(md5Sum, DIGITS_LOWER));
						if (dbInfo.md5.equals(md5download)) {
							UIHandler.showMessage(KEY_MSG_LOG_INFO, Common.getString(R.string.downloadInfoCRCOk));
							this.success = true;
							i = numUrl;
						} else {
							UIHandler.showMessage(KEY_MSG_LOG_WARN, Common.getString(R.string.downloadWarnCRCKo));
							Log.w(TAG, "md5 expected: " + dbInfo.md5);
							Log.w(TAG, "md5 compute: " + md5download);
							if (i + 1 < numUrl) {
								UIHandler.showMessage(KEY_MSG_LOG_WARN,
										Common.getString(R.string.downloadWarnCRCKoNextUrl));
							} else {
								this.success = true;
							}
						}
					} else {
						this.success = true;
						i = numUrl;
					}
				} catch (UnknownHostException uhe) {
					UIHandler.showMessage(KEY_MSG_LOG_ERROR, "UnknownHostException: " + uhe.getMessage());
					this.resetDownloadSize();
				} catch (NullPointerException npe) {
					// If we do not have some rights (like writing on the
					// sdcard, or downloading), a null pointer may be thrown
					UIHandler.showMessage(KEY_MSG_LOG_ERROR, "NullPointerException: " + npe.getMessage());
					this.resetDownloadSize();
				} catch (EOFException eofe) {
					UIHandler.showMessage(KEY_MSG_LOG_ERROR, "EOFException: " + eofe.getMessage());
					this.resetDownloadSize();
				} catch (IOException ioe) {
					UIHandler.showMessage(KEY_MSG_LOG_ERROR, "IOException: " + ioe.getMessage());
					this.resetDownloadSize();
				} catch (OutOfMemoryError oom) {
					UIHandler.showMessage(KEY_MSG_LOG_ERROR,
							Common.getString(R.string.downloadErrorOOM, oom.getMessage()));
					this.resetDownloadSize();

				} finally {

					try {
						if (null != gzIs)
							gzIs.close();
					} catch (IOException ioe) {
						UIHandler.showMessage(KEY_MSG_LOG_ERROR, "IOException: " + ioe.getMessage());
						this.resetDownloadSize();
					}

					try {
						if (null != fileOs)
							fileOs.close();
					} catch (IOException ioe) {
						UIHandler.showMessage(KEY_MSG_LOG_ERROR, "IOException: " + ioe.getMessage());
						this.resetDownloadSize();
					}

					if (null != connection) {
						connection.disconnect();
						connection = null;
					}
				}

			} // for

			if (success) {
				dbInfo.saveToPref();
			}

		} // run()

		/**
		 * <p>
		 * This method comes from Apache Commons Codec 1.7, and is license under
		 * the <a href="http://www.apache.org/licenses/LICENSE-2.0">Apache
		 * License Version 2.0</a><br/>
		 * Thanks to the The Apache Software Foundation.
		 * </p>
		 * Converts an array of bytes into an array of characters representing
		 * the hexadecimal values of each byte in order. The returned array will
		 * be double the length of the passed array, as it takes two characters
		 * to represent any given byte.
		 * 
		 * @param data
		 *            a byte[] to convert to Hex characters
		 * @param toDigits
		 *            the output alphabet
		 * @return A char[] containing hexadecimal characters
		 * @since 1.4
		 */
		private static char[] encodeHex(byte[] data, char[] toDigits) {
			int l = data.length;
			char[] out = new char[l << 1];
			// two characters form the hex value.
			for (int i = 0, j = 0; i < l; i++) {
				out[j++] = toDigits[(0xF0 & data[i]) >>> 4];
				out[j++] = toDigits[0x0F & data[i]];
			}
			return out;
		}

		private void resetDownloadSize() {
			synchronized (TAG) {
				this.downloadSize = -1;
				this.cIS = null;
			}
		}

		private long getDownloadSize() {
			return this.downloadSize;
		}

		private long getDownloadDoneInByte() {
			if (null != this.cIS) {
				return this.cIS.readCount;
			}
			return 0;
		}

	}

	/**
	 * Control the downloads, and update the progress bars
	 * 
	 */
	private static class Controller extends Thread {

		public Controller() {
			this.setName("Downloader Controller");
		}

		@Override
		public void run() {

			final int numDownloads = DBInfos.arrayDBInfos.length;
			boolean ok = true;

			// Calc total download size
			long totalDownloadSize = 0;
			for (int i = 0; i < numDownloads; i++) {
				final DBInfos dbInfo = DBInfos.arrayDBInfos[i];
				if (dbInfo.selected) {
					totalDownloadSize += dbInfo.gzSize;
				}
			}
			long totalDownloadDone = 0;

			for (int i = 0; i < numDownloads; i++) {
				final DBInfos dbInfo = DBInfos.arrayDBInfos[i];
				if (!dbInfo.selected) {
					continue;
				}

				long oldDownloadValue = 0;

				UIHandler.startDownload(i);

				Downloader downloader = new Downloader(dbInfo);
				downloader.start();

				boolean loop = true;

				while (loop) {
					try {
						super.sleep(DownloaderGzipDB.UPDATE_TIME);
					} catch (InterruptedException e) {
						loop = false;
					}
					if (DownloaderGzipDB.isPrematureThreadStop()) {
						return;
					}

					boolean running = downloader.isAlive();
					boolean success = downloader.success;

					if (running) {
						final long downloadSize = downloader.getDownloadSize();
						if (downloadSize > 0) {
							final long downloadDone = downloader.getDownloadDoneInByte();
							final int percent = (int) (downloadDone * 100 / downloadSize);
							final int totalPercent = (int) ((totalDownloadDone + downloadDone) * 100 / totalDownloadSize);
							final long speed = (downloadDone - oldDownloadValue) * SPEED_FACTOR;
							UIHandler.updateProgress(totalPercent, percent, speed);
							oldDownloadValue = downloadDone;
						} else {
							int totalPercent = (int) (totalDownloadDone * 100 / totalDownloadSize);
							UIHandler.updateProgress(totalPercent, 0, 0);
						}
					} else if (success) {
						loop = false;
						ok = ok & success;

						totalDownloadDone += dbInfo.gzSize;
						final int totalPercent = (int) (totalDownloadDone * 100 / totalDownloadSize);

						UIHandler.updateProgress(totalPercent, 100, 0L);
					} else {
						loop = false;
						ok = ok & success;
						UIHandler.endDownloadError(i);
					}
				}

				downloader = null;
				System.runFinalization();
				System.gc();
			}
			UIHandler.endDownload(ok);
		}
	}

	/**
	 * Handle controller and threads message, to be sent to the UI.
	 * 
	 */
	private static class UIHandler extends Handler {
		private DownloaderActivity uiActivity;
		private StringBuilder stringBuilder;

		public UIHandler(DownloaderActivity activity) {
			this.uiActivity = activity;
			this.stringBuilder = new StringBuilder(1024);
		}

		public void updateUI(DownloaderActivity activity) {
			this.uiActivity = activity;
			if (null != this.uiActivity) {
				TextView tv = (TextView) activity.findViewById(R.id.downloadTextViewLogs);
				tv.setText(Html.fromHtml(this.stringBuilder.toString()));
			}
		}

		private static void updateProgress(int percentDoneTotal, int percentDone, long speed) {
			synchronized (TAG) {
				if (null != DownloaderGzipDB.handler) {
					Message msg = Message.obtain();
					Bundle data = new Bundle();
					data.putInt(KEY_MSG_UPD_TOTAL, percentDoneTotal);
					data.putInt(KEY_MSG_PB_UPD_VAL, percentDone);
					data.putLong(KEY_MSG_DNWL_SPEED, speed);
					msg.setData(data);
					DownloaderGzipDB.handler.sendMessage(msg);
				}
			}
		}

		private static void showMessage(String level, String message) {
			synchronized (TAG) {
				if (null != DownloaderGzipDB.handler) {
					Message msg = Message.obtain();
					Bundle data = new Bundle();
					data.putString(level, message);
					msg.setData(data);
					DownloaderGzipDB.handler.sendMessage(msg);
				}
			}

		}

		private static void startDownload(int id) {
			synchronized (TAG) {
				if (null != DownloaderGzipDB.handler) {
					Message msg = Message.obtain();
					Bundle data = new Bundle();
					data.putInt(KEY_MSG_START_DWNL, id);
					msg.setData(data);
					DownloaderGzipDB.handler.sendMessage(msg);
				}
			}
		}

		private static void endDownload(boolean ok) {
			synchronized (TAG) {
				if (null != DownloaderGzipDB.handler) {
					// show the main menu if everything is OK
					Message msg = Message.obtain();
					Bundle data = new Bundle();
					data.putBoolean(KEY_MSG_MENU, ok);
					msg.setData(data);
					DownloaderGzipDB.handler.sendMessage(msg);
				}
			}
		}

		private static void endDownloadError(int id) {
			synchronized (TAG) {
				if (null != DownloaderGzipDB.handler) {
					Message msg = Message.obtain();
					Bundle data = new Bundle();
					data.putInt(KEY_MSG_PB_ERROR, id);
					msg.setData(data);
					DownloaderGzipDB.handler.sendMessage(msg);
				}
			}
		}

		@Override
		public void handleMessage(Message msg) {

			DownloaderActivity activity = this.uiActivity;
			if (null == activity) {
				// prey to not have a KEY_MSG_MENU when it happen
				return;
			}

			Bundle data = msg.getData();
			if (data.containsKey(KEY_MSG_UPD_TOTAL)) {
				final int progressTotal = data.getInt(KEY_MSG_UPD_TOTAL);
				final int progress = data.getInt(KEY_MSG_PB_UPD_VAL);
				final long speed = data.getLong(KEY_MSG_DNWL_SPEED);
				activity.updateDownloadProgressAndSpeed(progressTotal, progress, speed);

			} else if (data.containsKey(KEY_MSG_START_DWNL)) {
				final int id = data.getInt(KEY_MSG_START_DWNL);
				final TextView tv = (TextView) activity.findViewById(R.id.downloadTextViewDatabase);
				final String text = activity.getString(R.string.downloadDBStart);
				final String dbName = activity.getString(DBInfos.arrayDBInfos[id].strDBNameId);
				tv.setText(String.format(text, dbName));

			} else if (data.containsKey(KEY_MSG_LOG_INFO)) {
				TextView tv = (TextView) activity.findViewById(R.id.downloadTextViewLogs);
				final String message = data.getString(KEY_MSG_LOG_INFO);
				StringBuilder sb = this.stringBuilder;
				sb.append(message);
				sb.append("<br/>");
				tv.setText(Html.fromHtml(sb.toString()));
				Log.i(TAG, message);

			} else if (data.containsKey(KEY_MSG_LOG_ERROR)) {
				TextView tv = (TextView) activity.findViewById(R.id.downloadTextViewLogs);
				final String message = data.getString(KEY_MSG_LOG_ERROR);
				StringBuilder sb = this.stringBuilder;
				sb.append("<font color=\"#C11B17\">");
				sb.append(message);
				sb.append("</font><br/>");
				tv.setText(Html.fromHtml(sb.toString()));
				Log.e(TAG, "" + message);

			} else if (data.containsKey(KEY_MSG_LOG_WARN)) {
				TextView tv = (TextView) activity.findViewById(R.id.downloadTextViewLogs);
				final String message = data.getString(KEY_MSG_LOG_WARN);
				StringBuilder sb = this.stringBuilder;
				sb.append("<font color=\"#FDD017\">");
				sb.append(message);
				sb.append("</font><br/>");
				tv.setText(Html.fromHtml(sb.toString()));
				Log.w(TAG, message);

			} else if (data.containsKey(KEY_MSG_PB_ERROR)) {
				activity.updateDownloadProgressAndSpeed(100, 100, -1);
			} else if (data.containsKey(KEY_MSG_MENU)) {
				boolean ok = data.getBoolean(KEY_MSG_MENU);
				if (ok) {
					DownloaderGzipDB.controller = null;
					DownloaderGzipDB.handler = null;
					this.stringBuilder = null;

					activity.downloadComplete();
				}
			}
		}
	}

	public static boolean isDownloading() {
		synchronized (TAG) {
			return (DownloaderGzipDB.controller != null);
		}
	}

	/**
	 * Launch the download of files
	 * 
	 * @param files
	 *            the filenames.
	 */
	public static void download(DownloaderActivity uiActivity) {

		synchronized (TAG) {
			if (null == DownloaderGzipDB.controller) {
				DownloaderGzipDB.prematureStop = false;
				DownloaderGzipDB.handler = new UIHandler(uiActivity);
				DownloaderGzipDB.controller = new Controller();
				DownloaderGzipDB.controller.start();
			}
		}
	}

	public static void updateUIActivity(DownloaderActivity uiActivity) {
		synchronized (TAG) {
			if (null != DownloaderGzipDB.handler) {
				DownloaderGzipDB.handler.updateUI(uiActivity);
			}
		}
	}

	public static void prematureThreadStop() {
		synchronized (TAG) {
			DownloaderGzipDB.prematureStop = true;
			DownloaderGzipDB.controller = null;
			DownloaderGzipDB.handler = null;
			System.runFinalization();
			System.gc();
		}
	}

	private static boolean isPrematureThreadStop() {
		synchronized (TAG) {
			return DownloaderGzipDB.prematureStop;
		}
	}

}
