/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.kanjidb;

import java.io.File;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

import org.tmatesoft.sqljet.core.SqlJetEncoding;
import org.tmatesoft.sqljet.core.SqlJetException;
import org.tmatesoft.sqljet.core.SqlJetTransactionMode;
import org.tmatesoft.sqljet.core.table.ISqlJetTable;
import org.tmatesoft.sqljet.core.table.SqlJetDb;
import org.vono.narau.kanjidb.Kanji.Element;

/**
 * TODO: put in common some part with JMdict (like TABLE_DBINFO)
 * 
 */
public class KanjiDBWriter {
	private static final String KEY_DB_VERSION = "version";
	private static final int VALUE_DB_VERSION = 3;
	private static final String KEY_DB_CREATION_DATE = "creation_date";
	private static final String VALUE_CREATION_DATE = String.format("%1$tY%1$tm%1$td", Calendar.getInstance());

	// KanjiDict
	private static final String TABLE_DBINFO = "db_info";
	private static final String TABLE_DICT_KANJI = "kdict_kanji";
	private static final String TABLE_DICT_LANG_MEANINGS = "kdict_meanings_%1$s";
	private static final String TABLE_DICT_INDEX = "kdict_index";
	private static final String TABLE_DICT_REF = "kdict_ref";
	private static final String TABLE_DICT_LANGS = "kdict_langs";

	// KanjiVG
	private static final String TABLE_KVG_GROUP = "kvg_group";
	private static final String TABLE_KVG_PATH = "kvg_path";
	private static final String TABLE_KVG_RAD = "kvg_radicals";
	private static final String TABLE_KVG_POS = "kvg_positions";
	private static final String VIEW_KVG_SCHOOL = "kvg_school";

	private static final String CREATE_TABLE = "CREATE TABLE ";

	private static final String CREATE_TABLE_DB_INFO = CREATE_TABLE + TABLE_DBINFO
			+ " (key TEXT PRIMARY KEY, value_str TEXT, value_int INTEGER);";

	/**
	 * The kanji table, merge the primary group. I took only the 2 first more
	 * used elements of the main group:<br/>
	 * <ul>
	 * <li>6404 kvg:element</li>
	 * <li>38 kvg:original</li>
	 * <li>3 kvg:partial</li>
	 * <li>1 kvg:phon</li>
	 * <li>4 kvg:position</li>
	 * <li>247 kvg:radical</li>
	 * <li>1 kvg:radicalForm</li>
	 * <li>31 kvg:variant</li>
	 * </ul>
	 * Unix cmd:
	 * 
	 * <pre>grep '^&lt;g id="kvg:....."' kanjivg.xml | sed 's/&lt;g
	 * id="kvg:....."//' | sed 's/="[^"]*"//g' | sed 's/&gt;//g' | sed 's/ /\n/g' |
	 * sort | uniq -c</pre>
	 */
	private static final String CREATE_TABLE_DICT_KANJI = CREATE_TABLE + TABLE_DICT_KANJI
			+ " (codepoint INTEGER PRIMARY KEY," + "grade INTEGER," + "frequency INTEGER," + "jlpt INTEGER,"
			+ "stroke_count INTEGER, reading_on TEXT, reading_kun TEXT);";

	private static final String CREATE_TABLE_DICT_LANG_MEANINGS = CREATE_TABLE + TABLE_DICT_LANG_MEANINGS
			+ " (codepoint INTEGER NOT NULL," + "meanings TEXT NOT NULL)"
			+ " CONSTRAINT meanings_%1$s_pk PRIMARY_KEY(codepoint);";

	private static final String CREATE_TABLE_DICT_LANGS = CREATE_TABLE + TABLE_DICT_LANGS
			+ "(lang TEXT PRIMARY KEY, nb_entries INTEGER, table_name TEXT)";

	/**
	 * This element contains the index numbers and similar unstructured
	 * information such as page numbers in a number of published dictionaries,
	 * and instructional books on kanji.
	 */
	private static final String CREATE_TABLE_DICT_IDX = CREATE_TABLE + TABLE_DICT_INDEX
			+ " (codepoint INTEGER, dictionary INTEGER, index_number TEXT)"
			+ " CONSTRAINT dict_idx_pk PRIMARY_KEY (codepoint, dictionary);";

	private static final String CREATE_TABLE_DICT_REF = CREATE_TABLE + TABLE_DICT_REF
			+ " (id INTEGER, name TEXT, label TEXT);";

	/**
	 * <ul>
	 * <li>5 kvg:radicalForm</li>
	 * <li>7 kvg:tradForm</li>
	 * <li>19 kvg:number</li>
	 * <li>468 kvg:partial</li>
	 * <li>1361 kvg:phon</li>
	 * <li>3425 kvg:part</li>
	 * <li>4487 kvg:original</li>
	 * <li>5892 kvg:variant</li>
	 * <li>35906 kvg:element</li>
	 * <li>7040 kvg:radical</li>
	 * <li>25228 kvg:position</li>
	 * </ul>
	 * 
	 * <pre>
	 * grep -E '<g id="kvg:.....-g[0-9]+"' kanjivg.xml | sed -E 's/.*<g id="kvg:.....-g[0-9]+"//'| sed 's/="[^"]*"//g' | sed 's/>//g' | sed 's/ /\n/g' | sort | uniq -c
	 * </pre>
	 */
	private static final String CREATE_TABLE_KVG_GROUP = CREATE_TABLE
			+ TABLE_KVG_GROUP
			+ " (codepoint INTEGER, id INTEGER, parent INTEGER NOT NULL, element INTEGER, position INTEGER, radical INTEGER, original INTEGER)"
			+ " CONSTRAINT group_pk PRIMARY_KEY (codepoint, id);";
	private static final String CREATE_TABLE_KVG_PATH = CREATE_TABLE + TABLE_KVG_PATH
			+ " (codepoint INTEGER NOT NULL, parent INTEGER NOT NULL, type TEXT, draw TEXT NOT NULL)"
			+ " CONSTRAINT path_pk PRIMARY_KEY (codepoint, parent);";
	/**
	 * radical: 'r'+id ("bottom", "kamae", "kamae1", "kamae2", "left", "nyo",
	 * "right", "tare", "top")
	 */
	private static final String CREATE_TABLE_KVG_RAD = CREATE_TABLE + TABLE_KVG_RAD
			+ " (id INTEGER NOT NULL, value TEXT NOT NULL) " + "CONSTRAINT " + TABLE_KVG_RAD + "_pk PRIMARY_KEY(id);";
	/**
	 * position: 'p'+id ("general", "nelson", "tradit")<br/>
	 */
	private static final String CREATE_TABLE_KVG_POS = CREATE_TABLE + TABLE_KVG_POS
			+ " (id INTEGER  NOT NULL, value TEXT NOT NULL) " + "CONSTRAINT " + TABLE_KVG_POS + "_pk PRIMARY_KEY(id);";

	private static final String CREATE_VIEW_KVG_SCHOOL = "CREATE VIEW " + VIEW_KVG_SCHOOL + " AS " + "SELECT DISTINCT "
			+ TABLE_DICT_KANJI + ".codepoint AS codepoint, " + TABLE_DICT_KANJI + ".jlpt AS jlpt, " + TABLE_DICT_KANJI
			+ ".grade AS grade " + "FROM " + TABLE_DICT_KANJI + "," + TABLE_KVG_PATH + " " + "WHERE "
			+ TABLE_DICT_KANJI + ".codepoint=" + TABLE_KVG_PATH + ".codepoint";
	// and (" + TABLE_DICT_KANJI+ ".jlpt is not null or
	// " + TABLE_DICT_KANJI + ".grade is not null)

	private static final String CREATE_INDEX = "CREATE INDEX idx_%1$s_id ON %1$s (codepoint);";

	private SqlJetDb db;
	private ArrayList<Kanji> listKanji;
	/**
	 * Ugly hack to get the dict ref id in the database
	 */
	private HashMap<String, Integer> dictRef;

	public KanjiDBWriter(ArrayList<Kanji> kanjis) {
		this.listKanji = kanjis;
		this.dictRef = null;
	}

	public void createDB(File dbFile) throws SqlJetException {
		System.out.println("Creating " + dbFile.getName() + "...");

		this.db = new SqlJetDb(dbFile, true);
		this.db.open();
		this.db.getOptions().setAutovacuum(true);
		this.db.getOptions().setEncoding(SqlJetEncoding.UTF8);
		this.db.beginTransaction(SqlJetTransactionMode.WRITE);

		try {
			this.createTables();

			ISqlJetTable table = this.db.getTable(TABLE_DBINFO);
			table.insert(KEY_DB_VERSION, null, VALUE_DB_VERSION);
			table.insert(KEY_DB_CREATION_DATE, VALUE_CREATION_DATE, null);

			table = this.db.getTable(TABLE_DICT_REF);
			this.writeDictRef(table);

			ISqlJetTable tableKanji = this.db.getTable(TABLE_DICT_KANJI);
			ISqlJetTable tableGroup = this.db.getTable(TABLE_KVG_GROUP);
			ISqlJetTable tablePath = this.db.getTable(TABLE_KVG_PATH);
			ISqlJetTable tableDictIdx = this.db.getTable(TABLE_DICT_INDEX);

			final HashMap<String, ISqlJetTable> tablesMeanings = new HashMap<String, ISqlJetTable>(4);
			final HashMap<String, Integer> meaningCount = new HashMap<String, Integer>(4);

			final ArrayList<Kanji> lList = this.listKanji;
			final int count = lList.size();
			for (int i = 0; i < count; i++) {

				final Kanji kanji = lList.get(i);
				final int codepoint = kanji.codepoint;

				tableKanji.insert(codepoint, kanji.grade, kanji.frequency, kanji.jlpt, kanji.strokeCount,
						kanji.readingsOn, kanji.readingsKun);

				this.writeElement(tableGroup, tablePath, codepoint, -1, kanji.group);

				if (null != kanji.dicNumber) {
					this.writeDictIndex(tableDictIdx, codepoint, kanji.dicNumber);
				}

				if (null != kanji.meanings) {
					Set<Entry<String, String>> meanings = kanji.meanings.entrySet();
					for (Entry<String, String> meaning : meanings) {
						final String lang = meaning.getKey();
						ISqlJetTable tableMeanings = tablesMeanings.get(lang);
						if (null == tableMeanings) {
							this.db.createTable(String.format(CREATE_TABLE_DICT_LANG_MEANINGS, lang));
							tableMeanings = this.db.getTable(String.format(TABLE_DICT_LANG_MEANINGS, lang));
							tablesMeanings.put(lang, tableMeanings);
							meaningCount.put(lang, 1);
						} else {
							// Not optimise, but for this program, it is not
							// really important
							meaningCount.put(lang, meaningCount.get(lang) + 1);
						}
						tableMeanings.insert(codepoint, meaning.getValue());
					}
				}
			}

			final ISqlJetTable tableDictLangs = this.db.getTable(TABLE_DICT_LANGS);
			final Set<Entry<String, Integer>> setLangsCount = meaningCount.entrySet();
			for (Entry<String, Integer> entry : setLangsCount) {
				final String lang = entry.getKey();
				tableDictLangs.insert(lang, entry.getValue(), String.format(TABLE_DICT_LANG_MEANINGS, lang));
			}

			final ISqlJetTable tableKPos = this.db.getTable(TABLE_KVG_POS);
			final int lenPos = Kanji.POSITIONS.length;
			for (int i = 0; i < lenPos; i++) {
				tableKPos.insert(i, Kanji.POSITIONS[i]);
			}
			final ISqlJetTable tableKRad = this.db.getTable(TABLE_KVG_RAD);
			final int lenRad = Kanji.RADICAL.length;
			for (int i = 0; i < lenRad; i++) {
				tableKRad.insert(i, Kanji.RADICAL[i]);
			}

			this.db.createIndex(String.format(CREATE_INDEX, TABLE_DICT_KANJI));
			this.db.createIndex(String.format(CREATE_INDEX, TABLE_KVG_GROUP));
			this.db.createIndex(String.format(CREATE_INDEX, TABLE_KVG_PATH));
			this.db.createIndex(String.format(CREATE_INDEX, TABLE_DICT_INDEX));

			for (String lang : meaningCount.keySet()) {
				final String tableName = String.format(TABLE_DICT_LANG_MEANINGS, lang);
				final String createIndexCmd = String.format(CREATE_INDEX, tableName);
				this.db.createIndex(createIndexCmd);
			}

			this.db.createView(CREATE_VIEW_KVG_SCHOOL);

		} finally {
			this.db.commit();
			this.db.close();
		}
	}

	private void writeElement(ISqlJetTable tableGroup, ISqlJetTable tablePath, int codepoint, int parentId,
			Kanji.Element element) throws SqlJetException {
		if (element instanceof Kanji.Group) {
			final Kanji.Group group = (Kanji.Group) element;
			tableGroup.insert(codepoint, group.id, parentId, group.element, group.position, group.radical,
					group.original);

			final List<Element> elements = group.elements;
			final int numElem = elements.size();
			if (0 == numElem) {
				System.err.println("WARN: codepoint " + Integer.toHexString(codepoint) + " group " + group.id
						+ " with no sub element ");
			}

			for (int i = 0; i < numElem; i++) {
				writeElement(tableGroup, tablePath, codepoint, group.id, elements.get(i));
			}
		} else if (element instanceof Kanji.Path) {
			Kanji.Path path = (Kanji.Path) element;
			tablePath.insert(codepoint, parentId, path.type, path.draw);
		}
	}

	private void writeDictIndex(ISqlJetTable tableDictIdx, int codepoint, HashMap<String, String> dicNumber)
			throws SqlJetException {
		for (Entry<String, String> es : dicNumber.entrySet()) {

			final Integer id = this.dictRef.get(es.getKey());
			final String indexNumber = es.getValue();
			tableDictIdx.insert(codepoint, id, indexNumber);

		}
	}

	private void writeDictRef(ISqlJetTable tableDictRef) throws SqlJetException {
		int pos = 0;
		this.dictRef = new HashMap<String, Integer>();

		for (String key : Kanji.dictNames.keySet()) {
			String val = Kanji.dictNames.get(key);
			tableDictRef.insert(pos, key, val);
			this.dictRef.put(key, pos);
			pos++;
		}
	}

	private void createTables() throws SqlJetException {
		this.db.createTable(CREATE_TABLE_DB_INFO);
		this.db.createTable(CREATE_TABLE_DICT_KANJI);
		this.db.createTable(CREATE_TABLE_DICT_IDX);
		this.db.createTable(CREATE_TABLE_DICT_REF);
		this.db.createTable(CREATE_TABLE_DICT_LANGS);
		this.db.createTable(CREATE_TABLE_KVG_GROUP);
		this.db.createTable(CREATE_TABLE_KVG_PATH);
		this.db.createTable(CREATE_TABLE_KVG_POS);
		this.db.createTable(CREATE_TABLE_KVG_RAD);

	}

}
